@@ -15,6 +15,14 @@ func setNoDelay(fd *netFD, noDelay bool) error {
 	return syscall.ENOPROTOOPT
 }
 
-func setKeepAlivePeriod(fd *netFD, d time.Duration) error {
+func setKeepAliveIdle(fd *netFD, d time.Duration) error {
+	return syscall.ENOPROTOOPT
+}
+
+func setKeepAliveInterval(fd *netFD, d time.Duration) error {
+	return syscall.ENOPROTOOPT
+}
+
+func setKeepAliveCount(fd *netFD, n int) error {
 	return syscall.ENOPROTOOPT
 }
