@@ -8,11 +8,11 @@ package os
 
 import "syscall"
 
-func ensurePidfd(sysAttr *syscall.SysProcAttr) *syscall.SysProcAttr {
-	return sysAttr
+func ensurePidfd(sysAttr *syscall.SysProcAttr) (*syscall.SysProcAttr, bool) {
+	return sysAttr, false
 }
 
-func getPidfd(_ *syscall.SysProcAttr) (uintptr, bool) {
+func getPidfd(_ *syscall.SysProcAttr, _ bool) (uintptr, bool) {
 	return 0, false
 }
 
@@ -20,8 +20,6 @@ func pidfdFind(_ int) (uintptr, error) {
 	return 0, syscall.ENOSYS
 }
 
-func (p *Process) pidfdRelease() {}
-
 func (_ *Process) pidfdWait() (*ProcessState, error) {
 	panic("unreachable")
 }
